# this script learns a decision tree classifier, 
# generate a DNF rule for the positive class (in the standard binary classification setting),
# negate the DNF to CNF


import sklearn.datasets
from sklearn import tree
import numpy as np
from sklearn.tree import _tree
from sklearn.model_selection import train_test_split, KFold
from sklearn import metrics 
# from fairness_comparison.fairness.data.objects.ProcessedData import ProcessedData
# from fairness_comparison.fairness.data.objects.Adult import Adult
import pandas as pd
import justicia.utils as utils
import os
from data.objects import titanic as titanic_   
from data.objects import adult as adult_
from data.objects import ricci as ricci_
# from fairness_comparison.fairness.data.objects import Ricci
from sklearn.metrics import roc_auc_score
from sklearn.model_selection import GridSearchCV



class dtWrapper():

    def __init__(self, DecisionTreeClassifier, data_features_original, sensitive_attributes, negate = False,  verbose = True):

        self.nodes = []
        self.paths = [] # a DNF formula, only capture when the leaf is 1
        self.tree_to_code(DecisionTreeClassifier, data_features_original, negate, verbose = verbose)

        self.num_attributes = len(self.nodes)
        self.attributes = [] # does not include variables related to sensitive attributes
        self.sensitive_attributes = []
        self.attribute_variable_map = {}
        self.variable_attribute_map = {}
        self.auxiliary_variables = []
        self.classifier = []
        
        self._do_map_from_node_to_attribute_index(sensitive_attributes, verbose)
        self._construct_CNF()


    def _do_map_from_node_to_attribute_index(self, sensitive_attributes, verbose):
        """  
        TODO make it more efficient
        """
        cnt = 1
        _sensitive_attributes = [_var for _group in sensitive_attributes for _var in _group]
        
        self.sensitive_attributes = [[] for _ in sensitive_attributes]
        # appearance of sensitive attributes
        _visited_sensitive_attributes = []
        for node in self.nodes:
            feature, threshold = node
            # if feature is one-hot encoded, it may contain specific prefix


            if(feature in _sensitive_attributes): # when sensitive feature is binary
                _visited_sensitive_attributes.append(feature)
                self.attribute_variable_map[(feature, threshold)] = cnt
                self.variable_attribute_map[cnt] = (feature, threshold)
                for i in range(len(sensitive_attributes)):
                    if(feature in sensitive_attributes[i]):
                        self.sensitive_attributes[i].append(cnt)
                        
    
                cnt += 1
            else:
                self.attribute_variable_map[(feature, threshold)] = cnt
                self.variable_attribute_map[cnt] = (feature, threshold)
                self.attributes.append(cnt)
                cnt += 1

        
        # It may happen that all sensitive attributes are not preset in the decision tree
        for _sensitive_attribute in _sensitive_attributes:
            if(_sensitive_attribute not in _visited_sensitive_attributes):
                self.attribute_variable_map[(_sensitive_attribute, 0.5)] = cnt
                self.variable_attribute_map[cnt] = (_sensitive_attribute, 0.5)
                for i in range(len(sensitive_attributes)):
                    if(_sensitive_attribute in sensitive_attributes[i]):
                        self.sensitive_attributes[i].append(cnt)
                self.num_attributes += 1
                cnt += 1
            # else:
            #     print(_sensitive_attribute, "with assigned variable" ,self.attribute_variable_map[(_sensitive_attribute, 0.5)], "is in the tree")
            
        self._max_attribute_index = cnt - 1

        if(verbose):
            print("\nattribute to variable map ->")
            print(self.attribute_variable_map)
            print("\nvarible to attribute map ->")
            print(self.variable_attribute_map)
            
        assert self.num_attributes == len(self.attribute_variable_map), "Error in mapping attributes to index"

    def _construct_CNF(self):
        """  
        Applying De Morgan's law
        """
        # first construct a DNF formula from paths leading to 0 of the decision trees and then negate it to construct a CNF
        # self.path is already in DNF for class = 0

        for path in self.paths:
            clause = []
            for (name, threshold, flag) in path:
                clause.append(self.attribute_variable_map[(name, threshold)]  * -1 * flag) # -1 is used to negate the variable
            self.classifier.append(clause)


    def compute_probability(self, data, verbose = False):
        """
        inputs:
            data is a dataframe
        outputs:
            probabilities of attributes: a dictionary
        """
        if(verbose):
            print("\nCalculated probability")
        probs = {}
        for key in self.attribute_variable_map:
            feature, threshold = key
            probs[self.attribute_variable_map[key]] = round((data[feature] <= threshold).mean(),3)
            if(probs[self.attribute_variable_map[key]] == 0):
                probs[self.attribute_variable_map[key]] = 0.001
            elif(probs[self.attribute_variable_map[key]] == 1):
                probs[self.attribute_variable_map[key]] = 0.999
            if(verbose):
                print(feature, "<=" ,threshold, " has probability: " , probs[self.attribute_variable_map[key]], " and assigned variable is:", self.attribute_variable_map[key])
        return probs


    def compute_pair_wise_correlation(self, data, probs, verbose = True):


        attribute_cnt = self._max_attribute_index
        constraints = []
        
        for key_1 in self.attribute_variable_map:
            # if(self.attribute_variable_map[key_1] not in self.attributes):
            #         continue
                
            for key_2 in self.attribute_variable_map:   
                if(key_1 == key_2):
                    continue
                
                # if(self.attribute_variable_map[key_2] not in self.attributes):
                #     continue
                

                feature_1, threshold_1 = key_1
                feature_2, threshold_2 = key_2
                
                attribute_cnt += 1
                self.attributes.append(attribute_cnt)
                constraints.append((attribute_cnt, self.attribute_variable_map[key_1], self.attribute_variable_map[key_2]))
                probs[attribute_cnt] = round(((data[feature_1] <= threshold_1) & (data[feature_2] <= threshold_2)).mean(),3)

                attribute_cnt += 1
                self.attributes.append(attribute_cnt)
                constraints.append((attribute_cnt, self.attribute_variable_map[key_1], -1 * self.attribute_variable_map[key_2]))
                probs[attribute_cnt] = round(((data[feature_1] <= threshold_1) & (data[feature_2] > threshold_2)).mean(),3)

                attribute_cnt += 1
                self.attributes.append(attribute_cnt)
                constraints.append((attribute_cnt, -1 * self.attribute_variable_map[key_1], self.attribute_variable_map[key_2]))
                probs[attribute_cnt] = round(((data[feature_1] > threshold_1) & (data[feature_2] <= threshold_2)).mean(),3)

                attribute_cnt += 1
                self.attributes.append(attribute_cnt)
                constraints.append((attribute_cnt, -1 * self.attribute_variable_map[key_1], -1 * self.attribute_variable_map[key_2]))
                probs[attribute_cnt] = round(((data[feature_1] > threshold_1) & (data[feature_2] > threshold_2)).mean(),3)


        # for constraint in correlation_constraints:
        #     self.formula += self._construct_clause([-1 * constraint[0], constraint[1]])
        #     self.formula += self._construct_clause([-1 * constraint[0], constraint[2]])
        #     # self.formula += self._construct_clause([constraint[0], -1 * constraint[1], -1 * constraint[2]])


            #     break
            # break

        # print(constraints)
        # print("\n")
        # print(probs)
        
        return probs, constraints


             

        





    def tree_to_code(self, tree, feature_names, negate, verbose = True):

        
        tree_ = tree.tree_
        feature_name = [
            feature_names[i] if i != _tree.TREE_UNDEFINED else "undefined!"
            for i in tree_.feature
        ]
        if(verbose):
            print("\nLearned tree -->\n")
            print("def tree({}):".format(", ".join(feature_names)))

        def recurse(node, depth, path):
            indent = "  " * depth
            if tree_.feature[node] != _tree.TREE_UNDEFINED:
                name = feature_name[node]
                threshold = tree_.threshold[node]
                if((name, threshold) not in self.nodes):
                    self.nodes.append((name,  threshold)) # stores the description of the node (a Boolean variable)
                if(verbose):
                    print("{}if {} <= {}:".format(indent, name, threshold))
                recurse(tree_.children_left[node], depth + 1, path + [(name,  threshold, 1)])
                if(verbose):
                    print("{}else:  # NEG if {} <= {}".format(indent, name, threshold))
                recurse(tree_.children_right[node], depth + 1, path + [(name, threshold, -1)])
            else:
                class_label = np.argmax(tree_.value[node][0])
                if(class_label == 0 and not negate):
                    self.paths.append(path)
                if(class_label == 1 and negate):
                    self.paths.append(path)
                
                if(verbose):
                    print("{}return {}".format(indent, np.argmax(tree_.value[node][0])))

        recurse(0, 1, [])

        
        if(verbose):
            print("\n\n\nnodes ->")
            for node in self.nodes:
                print(node)

            print("\n\n\nDNF paths ->")            
            for path in self.paths:
                print(path)
        

    


# supporting functions for learning decision trees

def _get_sensitive_attibutes(known_sensitive_features, features):
    """ 
    Return sensitive attributes in appropriate format
    """

    # Extract new names of sensitive attributes
    _sensitive_attributes = {} # it is a map because each entry contains all one-hot encoded variables
    for _column in features:
        if("_" in _column and _column.split("_")[0] in known_sensitive_features):
            if(_column.split("_")[0] not in _sensitive_attributes):
                _sensitive_attributes[_column.split("_")[0]] = [_column]
            else:
                _sensitive_attributes[_column.split("_")[0]].append(_column)
        elif(_column in known_sensitive_features):
            if(_column not in _sensitive_attributes):
                _sensitive_attributes[_column] = [_column]
            else:
                _sensitive_attributes[_column].append(_column)


    # Finally make a 2d list
    sensitive_attributes = []
    for key in _sensitive_attributes:
        sensitive_attributes.append(_sensitive_attributes[key])


    return sensitive_attributes
    

def init_iris():

    # dataset.data is a np matrix, 
    # dataset.target is a np array
    # dataset['features] is the list of features in the original dataset

    # prepare iris dataset for binary classification
    target = "target"
    dataset = sklearn.datasets.load_iris()
    dataset[target] = np.where(dataset[target]==2, 0, dataset[target])

    # get df
    dataset = utils.sklearn_to_df(dataset)

    index_of_sensitive_features = 0

    # discretize sensitive attributes
    data =  utils.get_discretized_df(dataset, columns_to_discretize=[dataset.columns.to_list()[index_of_sensitive_features]])

    # get X,y
    X = data.drop(['target'], axis=1)
    y = data['target']


    # one-hot
    X = utils.get_one_hot_encoded_df(X,X.columns.to_list())

    # split into train_test
    X_train, X_test, y_train, y_test =  train_test_split(X,y, test_size=0.3, random_state=0)


 
    
    # Extract new names of sensitive attributes
    _sensitive_attributes = {} # it is a map because each entry contains all one-hot encoded variables
    for _column in X_train.columns.to_list():
        if("_" in _column and _column.split("_")[0] in dataset.columns.to_list()[index_of_sensitive_features]):
            if(_column.split("_")[0] not in _sensitive_attributes):
                _sensitive_attributes[_column.split("_")[0]] = [_column]
            else:
                _sensitive_attributes[_column.split("_")[0]].append(_column)
        elif(_column in dataset.columns.to_list()[index_of_sensitive_features]):
            if(_column not in _sensitive_attributes):
                _sensitive_attributes[_column] = [_column]
            else:
                _sensitive_attributes[_column].append(_column)
    
    # Finally make a 2d list
    sensitive_attributes = []
    for key in _sensitive_attributes:
        sensitive_attributes.append(_sensitive_attributes[key])
        
        
    

    
    clf = tree.DecisionTreeClassifier()
    clf = clf.fit(X_train, y_train)
    
    predict_train = clf.predict(X_train)
    predict_test = clf.predict(X_test)

    
    print("Train accuracy:",metrics.accuracy_score(y_train, predict_train), "positive ratio: ",y_train.mean())
    print("Test accuracy:",metrics.accuracy_score(y_test, predict_test), "positive ratio: ",y_test.mean())
    print("Train set positive prediction",predict_train.mean())
    print("Test set positive prediction",predict_test.mean())
    
    
    return clf, X_train.columns.to_list(), sensitive_attributes, X_train, X_test





def init_synthetic():
    filename = "data/sample.csv"

    if( os.path.isfile(filename)): 
        dataframe = pd.read_csv(filename)
    else:

        cols = 4
        rows = 200

        matrix = np.random.randint(2, size=(rows, cols))
        dataframe = pd.DataFrame.from_records(matrix)
        dataframe.columns = ['col-' + str(i) for i in range(cols-1)] + ['target']
        dataframe.to_csv(filename, index=False)


    # get X,y
    X = dataframe.drop(['target'], axis=1)
    y = dataframe['target']


    # one-hot
    X = utils.get_one_hot_encoded_df(X,X.columns.to_list())

    # split into train_test
    X_train, X_test, y_train, y_test =  train_test_split(X,y, test_size=0.3, random_state=0)


    sensitive_attributes = _get_sensitive_attibutes(['col-0'],X_train.columns.to_list())

    clf = tree.DecisionTreeClassifier()
    clf = clf.fit(X_train, y_train)
    predict_train = clf.predict(X_train)
    predict_test = clf.predict(X_test)

    
    print("Train accuracy:",metrics.accuracy_score(y_train, predict_train), "positive ratio: ",y_train.mean())
    print("Test accuracy:",metrics.accuracy_score(y_test, predict_test), "positive ratio: ",y_test.mean())
    print("Train set positive prediction",predict_train.mean())
    print("Test set positive prediction",predict_test.mean())
    


    return clf, sensitive_attributes, X_train, X_test





def init(dataset, repaired=False, verbose = False, compute_equalized_odds = False):
    
    df = dataset.get_df(repaired=repaired)



    # get X,y
    X = df.drop(['target'], axis=1)
    y = df['target']



    # one-hot
    X = utils.get_one_hot_encoded_df(X,dataset.categorical_attributes, verbose=verbose)


    skf = KFold(n_splits=5, shuffle=True, random_state=10)
    skf.get_n_splits(X, y)

    X_trains = []
    y_trains = []
    X_tests = []
    y_tests = []
    clfs = []
    
    for train, test in skf.split(X, y):

        X_trains.append(X.iloc[train])
        y_trains.append(y.iloc[train])
        X_tests.append(X.iloc[test])
        y_tests.append(y.iloc[test])
    
    

        # apply gridsearch
        param_grid = {'max_depth': np.arange(2, 10)}
        grid_tree = GridSearchCV(tree.DecisionTreeClassifier(), param_grid)
        grid_tree.fit(X_trains[-1], y_trains[-1])
        tree_preds = grid_tree.predict_proba(X_tests[-1])[:, 1]
        tree_performance = roc_auc_score(y_tests[-1], tree_preds)
        clf = grid_tree.best_estimator_
        clfs.append(clf)


        # clf = tree.DecisionTreeClassifier()
        # clf = clf.fit(X_train, y_train)
        predict_train = clf.predict(X_trains[-1])
        predict_test = clf.predict(X_tests[-1])
    

        if(verbose):
            print("\nTraining result =>")
            print('dt: Area under the ROC curve = {}'.format(tree_performance))
            
            # getting the best models:
            print(grid_tree.best_params_)
            
            print("\nTrain accuracy:",metrics.accuracy_score(y_trains[-1], predict_train), "positive ratio: ",y_trains[-1].mean())
            print("Test accuracy:",metrics.accuracy_score(y_tests[-1], predict_test), "positive ratio: ",y_tests[-1].mean())
            print("Train set positive prediction",predict_train.mean())
            print("Test set positive prediction",predict_test.mean())


       




    if(compute_equalized_odds):
        return clfs, X_trains, X_tests, dataset.known_sensitive_attributes, y_trains, y_tests
    
    return clfs, X_trains, X_tests, dataset.known_sensitive_attributes
    




















    







import os
import subprocess
from pysat.formula import CNF
import numpy as np
from pysat.pb import *


class Fairness_verifier():

    def __init__(self, timeout = 400):
        self.num_variables = 0
        self.num_clauses = 0
        self.timeout = timeout
        self.execution_error = False
        pass

    def __str__(self):
        s = "\nFairness verifier ->\n-number of variables: %s\n-number is clauses %s" % (self.num_variables, self.num_clauses)
        return s

    def _apply_random_quantification(self, var, prob):
        return "r " + str(prob)+" "+str(var)+" 0\n"

    def _apply_exist_quantification(self, var):
        return "e " + str(var)+" 0\n"


    def invoke_SSAT_solver(self, filename, find_maximization = True, verbose = True):
        
        dir_path = os.path.dirname(os.path.realpath(filename))
        cmd = "timeout " + str(self.timeout) + " abc -c \"ssat "+ str(dir_path) + "/" +str(filename)+"\""
        try:
            cmd_output = subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            if(verbose):
                print(e.output)
            cmd_output = e.output
            self.execution_error = True
            
        lines = cmd_output.decode('utf-8').split("\n")

        # process output
        tight_upper_bound = False
        upper_bound = 0
        lower_bound = 0
        read_optimal_assignment = False
        if(verbose):
            print("\n")
        
        for line in lines:
            if(read_optimal_assignment):
                self.assignment_to_exists_variables = list(map(int, line[:-2].strip().split(" ")))
                if(verbose):
                    print("Learned assignment:",self.assignment_to_exists_variables)
                read_optimal_assignment = False
            if(line == "[INFO] Exactly solve the instance: upper bound is tight"):
                tight_upper_bound = True
            if(line.startswith("  > Upper bound =")):
                upper_bound = float(line.split(" ")[-1])
            if(line.startswith("  > Lower bound =")):
                lower_bound = float(line.split(" ")[-1])
            if(line.startswith("  > optimizing assignment to exist vars")):
                read_optimal_assignment = True 
                # read optimal assignment in the next iteration

        if(find_maximization):
            # print("Is upper bound tight? ", tight_upper_bound)
            if(verbose):
                if(self.instance == "Enum"):
                    print("upper bound", upper_bound)
                if(not tight_upper_bound):
                    print("lower bound", lower_bound)
                else:
                    print("upper bound is tight")
            self.upper_bound = upper_bound
            self.lower_bound = lower_bound
        else:
            if(verbose):
                print("upper bound", 1 - lower_bound)
                if(self.instance == "Enum"):
                    print("lower bound", 1 - upper_bound)
            self.upper_bound = 1 - lower_bound
            self.lower_bound = 1 - upper_bound

        if(verbose):
            print("\n===================================\n")

        os.system("rm " + str(dir_path) + "/" +str(filename))
        return self.execution_error
        


    def _construct_clause(self, vars):
        self.num_clauses += 1
        clause = ""
        for var in vars:
            clause += str(var) + " "

        clause += "0\n"
        return clause

    


    def _construct_header(self):
        return "p cnf " + str(self.num_variables) + " " + str(self.num_clauses) + "\n"

    def encoding_Enum_SSAT(self, classifier, attributes, sensitive_attributes, auxiliary_variables, probs, filename, sensitive_attributes_assignment, dependency_constraints = [], verbose=True):
        # classifier is assumed to be a boolean formula. It is a 2D list
        # attributes is a list of variables. The last attribute is the sensitive attribute
        # probs is a dictionary containing the i.i.d. probabilities of attributes
        # sensitive attribute is assumed to have the next index after attributes

        self.instance = "Enum"

        self.num_variables = len(attributes) + len([_var for _group in sensitive_attributes for _var in _group]) + len(auxiliary_variables)
        self.formula = ""

        # random quantification over non-sensitive attributes
        for attribute in attributes:
            self.formula += self._apply_random_quantification(
                attribute, probs[attribute])


        # the sensitive attribute is exist quantified
        for group in sensitive_attributes:
            for var in group:
                self.formula += self._apply_exist_quantification(var)

        
        
        # for each sensitive-attribute (one hot vector), at least one group must be True
        self._formula_for_equal_one_constraints = ""
        for _group in sensitive_attributes:
            if(len(_group)>1): # when categorical attribute is not Boolean
                equal_one_constraint = PBEnc.equals(lits=_group, weights=[1 for _ in _group],  bound=1, top_id=self.num_variables)
                for clause in equal_one_constraint.clauses:
                    self._formula_for_equal_one_constraints += self._construct_clause(clause)
                auxiliary_variables += [i for i in range(self.num_variables + 1, equal_one_constraint.nv + 1)]
                self.num_variables = max(equal_one_constraint.nv, self.num_variables)

        
        # other variables (auxiliary) are exist quantified
        for var in auxiliary_variables:
            self.formula += self._apply_exist_quantification(var)


        # clauses for the classifier
        for clause in classifier:
            self.formula += self._construct_clause(clause)


        # clauses for the dependency constraints
        for clause in dependency_constraints:
            self.formula += self._construct_clause(clause)


        
        # append previous constraints. TODO this is crucial as auxiliary variables are added after deriving the constraint.
        self.formula += self._formula_for_equal_one_constraints
        


        # specify group to measure fairness
        for _var in sensitive_attributes_assignment:
            self.formula += self._construct_clause([_var])


        
        
        # store in a file
        self.formula = self._construct_header() + self.formula[:-1]
        file = open(filename, "w")
        file.write(self.formula)
        file.close()

        # if(verbose):
        #     print("SSAT instance ->")
        #     print(self.formula)
    
    def encoding_Learn_SSAT(self, classifier, attributes, sensitive_attributes, auxiliary_variables, probs, filename, dependency_constraints = [], ignore_sensitive_attribute = None, verbose=True, find_maximization = True):
        """  
        Maximization-minimization algorithm
        """

        self.instance = "Learn"

        # TODO here we call twice: 1) get the sensitive feature with maximum favor, 2) get the sensitive feature with minimum favor
        # classifier is assumed to be a boolean formula. It is a 2D list
        # attributes, sensitive_attributes and auxiliary_variables is a list of variables
        # probs is the list of i.i.d. probabilities of attributes that are not sensitive

        self.num_variables = len(attributes) + len([_var for _group in sensitive_attributes for _var in _group]) + len(auxiliary_variables)
        self.formula = ""

        # the sensitive attribute is exist quantified
        for group in sensitive_attributes:
            if(ignore_sensitive_attribute == None or group != ignore_sensitive_attribute):
                for var in group:
                    self.formula += self._apply_exist_quantification(var)


        # random quantification over non-sensitive attributes
        if(len(attributes) > 0):
            for attribute in attributes:
                self.formula += self._apply_random_quantification(
                    attribute, probs[attribute])
        else:
            # dummy random variables
            self.formula += self._apply_random_quantification(
                    self.num_variables, 0.5)
            self.num_variables += 1

        if(ignore_sensitive_attribute != None):
            for var in ignore_sensitive_attribute:
                    self.formula += self._apply_exist_quantification(var)

        if(not find_maximization):
            _classifier = CNF(from_clauses=classifier)
            _negated_classifier = _classifier.negate(topv=self.num_variables)
            classifier = _negated_classifier.clauses
            # print(auxiliary_variables, self.num_variables, _negated_classifier.auxvars)
            auxiliary_variables += [i for i in range(self.num_variables + 1, _negated_classifier.nv + 1)]
            # print(auxiliary_variables)
            self.num_variables = max(_negated_classifier.nv, self.num_variables)

        
        # for each sensitive-attribute (one hot vector), at least one group must be True
        self._formula_for_equal_one_constraints = ""
        for _group in sensitive_attributes:
            if(len(_group)>1): # when categorical attribute is not Boolean
                equal_one_constraint = PBEnc.equals(lits=_group, weights=[1 for _ in _group],  bound=1, top_id=self.num_variables)
                for clause in equal_one_constraint.clauses:
                    self._formula_for_equal_one_constraints += self._construct_clause(clause)
                auxiliary_variables += [i for i in range(self.num_variables + 1, equal_one_constraint.nv + 1)]
                self.num_variables = max(equal_one_constraint.nv, self.num_variables)

        

        # other variables (auxiliary) are exist quantified
        for var in auxiliary_variables:
            self.formula += self._apply_exist_quantification(var)

        # clauses for the classifier
        for clause in classifier:
            self.formula += self._construct_clause(clause)

        # clauses for the dependency constraints
        for clause in dependency_constraints:
            self.formula += self._construct_clause(clause)

        self.formula += self._formula_for_equal_one_constraints


                

        # store in a file
        self.formula = self._construct_header() + self.formula[:-1]
        file = open(filename, "w")
        file.write(self.formula)
        file.close()
        
        # if(verbose):
        #     print("SSAT instance ->")
        #     print(self.formula)

        


    



from pyrulelearn.imli import imli
import justicia.utils as utils
from sklearn.model_selection import train_test_split, StratifiedKFold, KFold
from sklearn import metrics
import numpy as np


def init(dataset, repaired=False, verbose = False, negated = False, compute_equalized_odds = False):

    df = dataset.get_df(repaired=repaired)

    # discretize
    data =  utils.get_discretized_df(df, columns_to_discretize=dataset.continuous_attributes, verbose=verbose)

    # get X,y
    X = data.drop(['target'], axis=1)
    y = data['target']


    # one-hot
    X = utils.get_one_hot_encoded_df(X,X.columns.to_list(), verbose=verbose)


    skf = KFold(n_splits=5, shuffle=True, random_state=10)
    skf.get_n_splits(X, y)

    X_trains = []
    y_trains = []
    X_tests = []
    y_tests = []
    clfs = []
    clf_negs = []

    for train, test in skf.split(X, y):

        X_trains.append(X.iloc[train])
        y_trains.append(y.iloc[train])
        X_tests.append(X.iloc[test])
        y_tests.append(y.iloc[test])

        

        #  For linear classifier, we use Logistic regression model of sklearn
        clf = imli(num_clause=2, data_fidelity=10, work_dir="data/", rule_type="CNF", verbose=False)
            
        clf.fit(X_trains[-1].values, y_trains[-1].values)


        clf_neg = None
        if(negated):
            clf_neg = imli(num_clause=2, data_fidelity=10, work_dir="data/", rule_type="CNF", verbose=False)
            clf_neg.fit(X_trains[-1].values, np.array([1 - val for val in y_trains[-1].values]))


        clfs.append(clf)
        clf_negs.append(clf_neg)


        if(verbose):
            print("\nFeatures: ", X_trains[-1].columns.to_list())
            print("Number of features:", len(X_trains[-1].columns.to_list()))
            print("\nlearned rule:")
            print(clf.get_rule(X_trains[-1].columns.to_list()))
        
        
            
        if(verbose):
            print("\nTrain Accuracy Score: ", metrics.accuracy_score(clf.predict(X_trains[-1].values, y_trains[-1].values),y_trains[-1].values) , "positive ratio: ",y_trains[-1].mean())
            print("Test Accuracy Score: ", metrics.accuracy_score(clf.predict(X_tests[-1].values, y_tests[-1].values),y_tests[-1].values), "positive ratio: ",y_tests[-1].mean())
        
    if(compute_equalized_odds):
        return clfs, clf_negs, X_trains, X_tests, dataset.known_sensitive_attributes, y_trains, y_tests
       
    return clfs, clf_negs, X_trains, X_tests, dataset.known_sensitive_attributes




import pandas as pd
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.model_selection import train_test_split
from sklearn.datasets import load_iris 
from feature_engine import discretisers as dsc
from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import KFold
from sklearn.tree import _tree
from scipy.stats import norm

def get_population_model_fairsquare_format(data,sensitive_attribute):

    divide = True

    s = 'def popModel():\n'
    mean = data[sensitive_attribute].mean()
    mean = round(mean,7)
    s += "\t" + sensitive_attribute + " = step([(0,0.5," + str(round(1-mean,7)) + "), (0.5,1," + str(mean) + ")])\n"
    
    if(divide):
        s += "\tif " + str(sensitive_attribute)  +" < 0.5:\n" 
        for column in data.columns:
            if(column == sensitive_attribute):
                continue
            if(len(data[column].unique()) == 2):
                mean = data[data[sensitive_attribute] < 0.5][column].mean()
                mean = round(mean,7)
                s += "\t\t" + column + " = step([(0,0.5," + str(round(1-mean,7)) + "), (0.5,1," + str(mean) + ")])\n"
            else:    
                mu, std = norm.fit(data[data[sensitive_attribute] < 0.5] [column])
                mu = round(mu,7)
                std = round(std,7)
                s += "\t\t" + column + " = gaussian(" + str(mu) + "," + str(std) + ")\n"

        s += "\telse:\n"
        for column in data.columns:
            if(column == sensitive_attribute):
                continue
            if(len(data[column].unique()) == 2):
                mean = data[data[sensitive_attribute] >= 0.5][column].mean()
                mean = round(mean,7)
                s += "\t\t" + column + " = step([(0,0.5," + str(round(1-mean,7)) + "), (0.5,1," + str(mean) + ")])\n"
            else:    
                mu, std = norm.fit(data[data[sensitive_attribute] >= 0.5] [column])
                mu = round(mu,7)
                std = round(std,7)
                s += "\t\t" + column + " = gaussian(" + str(mu) + "," + str(std) + ")\n"
    else:
        for column in data.columns:
            if(column == sensitive_attribute):
                continue
            if(len(data[column].unique()) == 2):
                mean = data[column].mean()
                mean = round(mean,7)
                s += "\t" + column + " = step([(0,0.5," + str(round(1-mean,7)) + "), (0.5,1," + str(mean) + ")])\n"
            else:    
                mu, std = norm.fit(data[column])
                mu = round(mu,7)
                std = round(std,7)
                s += "\t" + column + " = gaussian(" + str(mu) + "," + str(std) + ")\n"
    s += "\tsensitiveAttribute(" + sensitive_attribute + " < 0.5)\n\n\n\n"
    return s


def get_population_model_verifair_format(data,sensitive_attribute):

    divide = True

    s = "from ..benchmarks.fairsquare.helper import *\n\n\ndef sample(flag):\n\n"

    mean = data[sensitive_attribute].mean()
    mean = round(mean,7)
    s += "\t" + sensitive_attribute + " = step([(0,1," + str(round(1-mean,7)) + "), (1,2," + str(mean) + ")])\n"
    
    if(divide):
        s += "\tif " + str(sensitive_attribute)  +" < 0.5:\n" 
        for column in data.columns:
            if(column == sensitive_attribute):
                continue
            if(len(data[column].unique()) == 2):
                mean = data[data[sensitive_attribute] < 0.5][column].mean()
                mean = round(mean,7)
                s += "\t\t" + column + " = step([(0,1," + str(round(1-mean,7)) + "), (1,2," + str(mean) + ")])\n"
            else:    
                mu, std = norm.fit(data[data[sensitive_attribute] < 0.5] [column])
                mu = round(mu,7)
                std = round(std,7)
                s += "\t\t" + column + " = gaussian(" + str(mu) + "," + str(std) + ")\n"

        s += "\telse:\n"
        for column in data.columns:
            if(column == sensitive_attribute):
                continue
            if(len(data[column].unique()) == 2):
                mean = data[data[sensitive_attribute] >= 0.5][column].mean()
                mean = round(mean,7)
                s += "\t\t" + column + " = step([(0,1," + str(round(1-mean,7)) + "), (1,2," + str(mean) + ")])\n"
            else:    
                mu, std = norm.fit(data[data[sensitive_attribute] >= 0.5] [column])
                mu = round(mu,7)
                std = round(std,7)
                s += "\t\t" + column + " = gaussian(" + str(mu) + "," + str(std) + ")\n"
    else:
        for column in data.columns:
            if(column == sensitive_attribute):
                continue
            if(len(data[column].unique()) == 2):
                mean = data[column].mean()
                mean = round(mean,7)
                s += "\t" + column + " = step([(0,1," + str(round(1-mean,7)) + "), (1,2," + str(mean) + ")])\n"
            else:    
                mu, std = norm.fit(data[column])
                mu = round(mu,7)
                std = round(std,7)
                s += "\t" + column + " = gaussian(" + str(mu) + "," + str(std) + ")\n"
    s += "\tsensitiveAttribute(" + sensitive_attribute + " < 1, flag)\n\n\n\n"
    return s

def attribute_raname_fairsquare_format(data):
    attributes = [] 
    for column in data.columns:
        column = column.strip().replace(" ", "_")
        column = column.replace("-", "_")
        column = column.replace("=", "_eq_")
        column = column.replace(">", "_g_")
        column = column.replace("<", "_l_")
        column = column.replace("(", "_lpar_")
        column = column.replace(")", "_rpar_")
        column = column.replace("/", "_or_")
        attributes.append(column)
    data.columns = attributes
    return data

def linear_model_fairsquare_format(clf, dataset_name, feature_names):
    coef = clf.coef_[0]
    bias = clf.intercept_[0]
    assert len(coef) == len(feature_names), "Error dimension of input attributes"
    s = "def F():\n\tt = "
    total_features = len(feature_names)
    for idx in range(total_features):
        s += str(round(coef[idx],7)) + " * " + feature_names[idx] + " + "
    s += str(round(bias,7))
    
    s += "\n\n\n"
    if(dataset_name == "titanic"  or dataset_name == "adult" or dataset_name == "german"):
        s += "\tfairnessTarget(t < 0)\n\n"
    else:
        s += "\tfairnessTarget(t >= 0)\n\n"
    
    return s


def linear_model_verifair_format(clf, dataset_name, feature_names):
    coef = clf.coef_[0]
    bias = clf.intercept_[0]
    assert len(coef) == len(feature_names), "Error dimension of input attributes"
    s = "\tt = "
    total_features = len(feature_names)
    for idx in range(total_features):
        s += str(round(coef[idx],7)) + " * " + feature_names[idx] + " + "
    s += str(round(bias,7))
    
    s += "\n\n\n"
    if(dataset_name == "titanic"  or dataset_name == "adult" or dataset_name == "german"):
        s += "\treturn int(t < 0)\n\tfairnessTarget(t < 0)\n\n"
    else:
        s += "\treturn int(t >= 0)\n\tfairnessTarget(t >= 0)\n\n"
    
    return s


def tree_to_code_fairsquare_format(tree, dataset_name, feature_names):

        
    tree_ = tree.tree_
    feature_name = [
        feature_names[i] if i != _tree.TREE_UNDEFINED else "undefined!"
        for i in tree_.feature
    ]
    s = "def tree({}):".format(", ".join(feature_names)) + "\n\n"
    s = "def F():\n"
    # print("\nLearned tree -->\n")
    # print("def tree({}):".format(", ".join(feature_names)))

    def recurse(node, depth, s):
        indent = "\t" * depth
        if tree_.feature[node] != _tree.TREE_UNDEFINED:
            name = feature_name[node]
            threshold = round(tree_.threshold[node],7)
            s = s + "{}if {} <= {}:".format(indent, name, threshold) + "\n"
            # print("{}if {} <= {}:".format(indent, name, threshold))
            s = recurse(tree_.children_left[node], depth + 1, s)
            s = s + "{}else:".format(indent) + "\n"
            # print("{}else:".format(indent))
            s = recurse(tree_.children_right[node], depth + 1, s)
        else:
            s = s + "{}t = {}".format(indent, np.argmax(tree_.value[node][0])) + "\n"
            # print("{}return {}".format(indent, np.argmax(tree_.value[node][0])))
        
        return s


    s = recurse(0, 1, s)
    if(dataset_name == "titanic"  or dataset_name == "adult" or  dataset_name == "german"):
        s += "\tfairnessTarget(t < 0.5)\n\n"
    else:
        s += "\tfairnessTarget(t >= 0.5)\n\n"


    return s


def tree_to_code_verifair_format(tree, dataset_name, feature_names):

        
    tree_ = tree.tree_
    feature_name = [
        feature_names[i] if i != _tree.TREE_UNDEFINED else "undefined!"
        for i in tree_.feature
    ]
    s = "def tree({}):".format(", ".join(feature_names)) + "\n\n"
    s = ""
    # print("\nLearned tree -->\n")
    # print("def tree({}):".format(", ".join(feature_names)))

    def recurse(node, depth, s):
        indent = "\t" * depth
        if tree_.feature[node] != _tree.TREE_UNDEFINED:
            name = feature_name[node]
            threshold = round(tree_.threshold[node],7)
            s = s + "{}if {} <= {}:".format(indent, name, threshold) + "\n"
            # print("{}if {} <= {}:".format(indent, name, threshold))
            s = recurse(tree_.children_left[node], depth + 1, s)
            s = s + "{}else:".format(indent) + "\n"
            # print("{}else:".format(indent))
            s = recurse(tree_.children_right[node], depth + 1, s)
        else:
            s = s + "{}t = {}".format(indent, np.argmax(tree_.value[node][0])) + "\n"
            # print("{}return {}".format(indent, np.argmax(tree_.value[node][0])))
        
        return s


    s = recurse(0, 1, s)
    if(dataset_name == "titanic"  or dataset_name == "adult" or  dataset_name == "german"):
        s += "\treturn int(t < 0.5)\n\tfairnessTarget(t < 0.5)\n\n"
    else:
        s += "\treturn int(t > 0.5)\n\tfairnessTarget(t >= 0.5)\n\n"


    return s


def aif360_cv_split(dataset, num_of_splits, shuffle=False, seed=None):
        
        
        n = dataset.features.shape[0]
        
        
        kf = KFold(n_splits=num_of_splits, shuffle=shuffle, random_state=10)
        kf.get_n_splits(dataset.features, dataset.labels)
        
        
        
        folds_train = [dataset.copy() for _ in range(num_of_splits)]
        folds_test = [dataset.copy() for _ in range(num_of_splits)]
        cnt = 0
        for train, test in kf.split(dataset.features, dataset.labels):
            fold = folds_train[cnt]
            fold.features = dataset.features[train]
            fold.labels = dataset.labels[train]
            fold.scores = dataset.scores[train]
            fold.protected_attributes = dataset.protected_attributes[train]
            fold.instance_weights = dataset.instance_weights[train]
            fold.instance_names = list(map(str, np.array(dataset.instance_names)[train]))
            fold.metadata = fold.metadata.copy()
            fold.metadata.update({
                'transformer': '{}.split'.format(type(dataset).__name__),
                'params': {'num_of_splits': num_of_splits,
                           'shuffle': shuffle},
                'previous': [dataset]
            })

            fold = folds_test[cnt]
            fold.features = dataset.features[test]
            fold.labels = dataset.labels[test]
            fold.scores = dataset.scores[test]
            fold.protected_attributes = dataset.protected_attributes[test]
            fold.instance_weights = dataset.instance_weights[test]
            fold.instance_names = list(map(str, np.array(dataset.instance_names)[test]))
            fold.metadata = fold.metadata.copy()
            fold.metadata.update({
                'transformer': '{}.split'.format(type(dataset).__name__),
                'params': {'num_of_splits': num_of_splits,
                           'shuffle': shuffle},
                'previous': [dataset]
            })

            cnt += 1
        
        return folds_train, folds_test

def get_scaled_df(X):
    # scale the feature values 
    sc = StandardScaler()
    X = sc.fit_transform(X)
    return X


def sklearn_to_df(sklearn_dataset):
    """ 
    Convert sklearn dataset to a dataframe, the class-label is renamed to "target"
    """
    df = pd.DataFrame(sklearn_dataset.data, columns=sklearn_dataset.feature_names)
    df['target'] = pd.Series(sklearn_dataset.target)
    return df


def get_discretized_df(data, columns_to_discretize = None, verbose=False):
    """ 
    returns train_test_splitted and discretized df
    """

    
    if(columns_to_discretize is None):
        columns_to_discretize = data.columns.to_list()

    if(verbose):
        print("Applying discretization\nAttribute bins")
    for variable in columns_to_discretize:
        bins = min(4, len(data[variable].unique()))
        if(verbose):
            print(variable, bins)
            
        # set up the discretisation transformer
        disc  = dsc.EqualWidthDiscretiser(bins=bins, variables = [variable])
        
        # fit the transformer
        disc.fit(data)

        if(verbose):
            print(disc.binner_dict_)

        

        # transform the data
        data = disc.transform(data)
        if(verbose):
            print(data[variable].unique())
        
        
    return data

def get_one_hot_encoded_df(df, columns_to_one_hot, verbose = False):
    """  
    Apply one-hot encoding on categircal df and return the df
    """
    for column in columns_to_one_hot:
        if(column not in df.columns.to_list()):
            if(verbose):
                print(column, " is not considered in classification")
            continue 

        # Apply when there are more than two categories or the binary categories are string objects.
        unique_categories = df[column].unique()
        if(len(unique_categories) > 2):
            one_hot = pd.get_dummies(df[column])
            if(len(one_hot.columns)>1):
                one_hot.columns = [column + "_" + str(c) for c in one_hot.columns]
            else:
                one_hot.columns = [column for c in one_hot.columns]
            df = df.drop(column,axis = 1)
            df = df.join(one_hot)
        else:
            if(0 in unique_categories and 1 in unique_categories):
                continue
            df[column] = df[column].map({unique_categories[0]: 0, unique_categories[1]: 1})
            if(verbose):
                print("Applying following mapping on attribute", column, "=>", unique_categories[0], ":",  0, "|", unique_categories[1], ":", 1)
            
    return df


def get_statistics_from_df(data, known_sensitive_attributes, verbose=False):
    # return non-sensitive attributes, sensitive attributes, probabilities of different attributes

    
    probs = {}
    sensitive_attributes = [[] for _ in known_sensitive_attributes]
    attributes = [] # contains non-sensitive attributes
    attribute_variable_map = {}
    variable_attribute_map = {}

    _mean = data.mean() 
    column_info = {}
    for idx in range(len(data.columns)):
        
        column_info[idx] = ("Bin", data.columns[idx])
        probs[idx + 1] = round(_mean[data.columns[idx]], 3)
        assert 0 <= probs[idx + 1] and probs[idx + 1] <= 1, "Error in calculating probabilities"
        if(probs[idx + 1] == 0):
            probs[idx + 1] = 0.001
        elif(probs[idx + 1] == 1):
            probs[idx + 1] = 0.999

        # check for sensitive attributes
        _is_sensitive_attribute = False
        for group_idx in range(len(known_sensitive_attributes)):
            if(data.columns[idx].split("_")[0] in known_sensitive_attributes[group_idx]):
                sensitive_attributes[group_idx].append(idx + 1)
                _is_sensitive_attribute = True
                variable_attribute_map[idx + 1] = data.columns[idx]
                attribute_variable_map[data.columns[idx]] = idx + 1
                    
                break

            elif(data.columns[idx] in known_sensitive_attributes[group_idx]):
                sensitive_attributes[group_idx].append(idx + 1)
                variable_attribute_map[idx + 1] = data.columns[idx]
                attribute_variable_map[data.columns[idx]] = idx + 1
                _is_sensitive_attribute = True
                break

        # otherwise non-sensitive attributes 
        if(not _is_sensitive_attribute):
            attributes.append(idx + 1)
            attribute_variable_map[data.columns[idx]] = idx + 1
            variable_attribute_map[idx + 1] = data.columns[idx] 
        
    if(verbose):
        print("\nvariable to attribute map ->")    
        print(variable_attribute_map)
        print("\nattribute to variable map ->")
        print(attribute_variable_map)
        print("\n\n\n")
        
        
        
    return attributes, sensitive_attributes, probs, variable_attribute_map, column_info


def calculate_probs_linear_classifier_wrap(data, column_info):

    
    probs = {}
    _attributes_cnt = 1
    for idx in range(len(data.columns)):

        if(column_info[idx][0] == "Bin"):
            probs[_attributes_cnt] = round(data[data.columns[idx]].mean(), 3)
            if(probs[_attributes_cnt] == 0):
                probs[_attributes_cnt] = 0.001
            elif(probs[_attributes_cnt] == 1):
                probs[_attributes_cnt] = 0.999

            _attributes_cnt += 1
        
        else:
            bins, _binner_dict, _ = column_info[idx]
            
            # print()
            # print(data.columns[idx])
            # mu, sigma = norm.fit(data[data.columns[idx]])
            # print("mu ", mu, "sigma ", sigma)  
            # print("max:", data[data.columns[idx]].max() )
            # print("min", data[data.columns[idx]].min() ) 
            # print()

            for i in range(bins):
                if(i == bins - 1 ):
                    # for the last bin, include the max boundary
                    probs[_attributes_cnt] = round(((_binner_dict[i] <= data[data.columns[idx]]) & (data[data.columns[idx]] <= _binner_dict[i+1])).mean(),3)    
                else:
                    probs[_attributes_cnt] = round(((_binner_dict[i] <= data[data.columns[idx]]) & (data[data.columns[idx]] < _binner_dict[i+1])).mean(),3)
                
                # probs[_attributes_cnt] = round(((_binner_dict[i] <= data[data.columns[idx]]) & (data[data.columns[idx]] < _binner_dict[i+1])).mean(),3)
                if(probs[_attributes_cnt] == 0):
                    probs[_attributes_cnt] = 0.001
                elif(probs[_attributes_cnt] == 1):
                    probs[_attributes_cnt] = 0.999
                _attributes_cnt += 1
    
    

    return probs

from scipy.stats import moment

def n_weighted_moment(values, weights, n):

    assert n>0 & (values.shape == weights.shape)
    w_avg = np.average(values, weights = weights)
    w_var = np.sum(weights * (values - w_avg)**2)/np.sum(weights)

    if n==1:
        return w_avg
    elif n==2:
        return w_var
    else:
        w_std = np.sqrt(w_var)
        return np.sum(weights * ((values - w_avg)/w_std)**n)/np.sum(weights)

from sklearn.neighbors import KernelDensity
import matplotlib
import matplotlib.pyplot as plt
from sklearn.model_selection import GridSearchCV
import statsmodels.api as sm

def get_statistics_for_linear_classifier_wrap(data, weights, known_sensitive_attributes, discretizer = "equalWidth", verbose=False):
    
    """
    TODO Data contains either binary attributes or real-valued attributes. 
    If it is binary attributes, the probability calculation is trivial. 
    For real-valued attributes, we need to discretize it. 
    let a is the coefficient and x is the real-valued features. 
    now x has to be discretized to, say z1, z2 and z3. 
    We apply the following transformation

    ax = a * sum_i=1^n (t_i * z_i) where t_i is the mean value of the threshold that z_i is covering 
    """
    # return non-sensitive attributes, sensitive attributes, probabilities of different attributes

    calculated_weights = []
    probs = {}
    sensitive_attributes = [[] for _ in known_sensitive_attributes]
    attributes = [] # contains non-sensitive attributes
    column_info = {}
    attribute_variable_map = {}
    variable_attribute_map = {}
    
    _attributes_cnt = 1
    _discretized_attributes_group = []
    for idx in range(len(data.columns)):
        
        _num_uniques = len(data[data.columns[idx]].unique())
        # assert _num_uniques != 1, "Error:  " + data.columns[idx] + " feature contains single feature-value"
        if( _num_uniques <= 2):

            # for sensitive attributes, the code reaches here
            # check for sensitive attributes
            _is_sensitive_attribute = False
            for group_idx in range(len(known_sensitive_attributes)):
                if(data.columns[idx].split("_")[0] in known_sensitive_attributes[group_idx]):
                    sensitive_attributes[group_idx].append(_attributes_cnt)
                    _is_sensitive_attribute = True
                    variable_attribute_map[_attributes_cnt] = data.columns[idx]
                    attribute_variable_map[data.columns[idx]] = _attributes_cnt
                    break

                elif(data.columns[idx] in known_sensitive_attributes[group_idx]):
                    sensitive_attributes[group_idx].append(_attributes_cnt)
                    _is_sensitive_attribute = True
                    variable_attribute_map[_attributes_cnt] = data.columns[idx]
                    attribute_variable_map[data.columns[idx]] = _attributes_cnt
                    break

            # otherwise non-sensitive attributes 
            if(not _is_sensitive_attribute):
                attributes.append(_attributes_cnt)
                variable_attribute_map[_attributes_cnt] = data.columns[idx]
                attribute_variable_map[data.columns[idx]] = _attributes_cnt
                                

            # binary features
            calculated_weights.append(weights[idx])
            probs[_attributes_cnt] = round(data[data.columns[idx]].mean(), 3)
            _attributes_cnt += 1
            
            column_info[idx] = ("Bin", data.columns[idx])
        
        else:
            bins = min(4, _num_uniques)
            if(discretizer == "equalWidth"):

                # set up the discretisation transformer
                disc  = dsc.EqualWidthDiscretiser(bins=bins, variables = [data.columns[idx]])
                
                # fit the transformer
                disc.fit(data)
            elif(discretizer == "equalFreq"):
                disc = dsc.EqualFrequencyDiscretiser(q=bins, variables=[data.columns[idx]])
                # fit the transformer
                disc.fit(data)
                
            else:
                raise ValueError(discretizer + " not defined")

            # calculate for introduced variables
            _binner_dict = disc.binner_dict_[data.columns[idx]]
            
            # print()
            # print(data.columns[idx])
            # mu, sigma = norm.fit(data[data.columns[idx]])
            # print("mu ", mu, "sigma ", sigma)   
            # print()

            assert _binner_dict[0] == float("-inf"), "Error: discretization error"
            assert _binner_dict[-1] == float("inf"), "Error: discretization error"
            assert len(_binner_dict) == bins + 1, "Error: wrong number of bins in discretization"
            _binner_dict[0] = data[data.columns[idx]].min() 
            _binner_dict[-1] = data[data.columns[idx]].max() 
            # _binner_dict[0] = 0
            # _binner_dict[-1] = 1
            _discretized_attributes_group.append([]) 


            histogram_mean = []
            histogram_prob = []
            histogram_mean_data_dependent = []

            for i in range(bins):
                attributes.append(_attributes_cnt)
                _threshold = float((_binner_dict[i+1] + _binner_dict[i])/2.0)
                histogram_mean.append(_threshold)
                if(i == bins - 1 ):
                    # for the last bin, include the max boundary
                    variable_attribute_map[_attributes_cnt] = (data.columns[idx], ">=" ,_binner_dict[i], "<=", _binner_dict[i+1])
                    attribute_variable_map[(data.columns[idx], ">=" ,_binner_dict[i], "<=", _binner_dict[i+1])] = _attributes_cnt
                    probs[_attributes_cnt] = round(((_binner_dict[i] <= data[data.columns[idx]]) & (data[data.columns[idx]] <= _binner_dict[i+1])).mean(),3)    
                    histogram_mean_data_dependent.append(data[data.columns[idx]][(_binner_dict[i] <= data[data.columns[idx]]) & (data[data.columns[idx]] <= _binner_dict[i+1])].mean())
                else:
                    variable_attribute_map[_attributes_cnt] = (data.columns[idx], ">=" ,_binner_dict[i], "<", _binner_dict[i+1])
                    attribute_variable_map[(data.columns[idx], ">=" ,_binner_dict[i], "<", _binner_dict[i+1])] = _attributes_cnt
                    probs[_attributes_cnt] = round(((_binner_dict[i] <= data[data.columns[idx]]) & (data[data.columns[idx]] < _binner_dict[i+1])).mean(),3)
                    histogram_mean_data_dependent.append(data[data.columns[idx]][(_binner_dict[i] <= data[data.columns[idx]]) & (data[data.columns[idx]] < _binner_dict[i+1])].mean())
                histogram_prob.append(probs[_attributes_cnt])

                # sample mean within bin-->
                _threshold = histogram_mean_data_dependent[-1]
                calculated_weights.append(weights[idx] * _threshold) # multiplied by the mean
                
                _discretized_attributes_group[-1].append(_attributes_cnt)
                _attributes_cnt += 1

            # print()
            # print(data.columns[idx])
            # # print(data[data.columns[idx]].values)
            # x = data[data.columns[idx]].values
            # params = {'bandwidth': np.logspace(-1, 1, 20)}
            # grid = GridSearchCV(KernelDensity(), params, cv=5)
            # grid.fit(x.reshape(-1,1))
            # kde = grid.best_estimator_
            # # vals = np.linspace(x.min(), x.max(), 100*bins)
            # vals = np.array(_binner_dict)
            # density = np.exp(kde.score_samples(vals.reshape(-1,1)))

            # kde_sm = sm.nonparametric.KDEUnivariate(np.array(x))
            # kde_sm.fit()
                
            # closest_indices = []
            # for q in range(len(vals)):
            #     closest_indices.append(kde_sm.support.tolist().index(min(kde_sm.support, key=lambda x:abs(x-vals[q]))))
            
            # cdf = kde_sm.cdf
            # histogram_prob_kde = []
            # for q in range(len(closest_indices)-1):
            #     if(q == 0):
            #         histogram_prob_kde.append(cdf[closest_indices[q+1]])
            #     elif(q == len(closest_indices)-2):
            #         histogram_prob_kde.append(1 - cdf[closest_indices[q]])
            #     else:          
            #         histogram_prob_kde.append(cdf[closest_indices[q+1]] - cdf[closest_indices[q]])
            



            # # print(vals)
            # # print(density)
            # print("histogram_prob", histogram_prob)
            # print("histogram_prob_kde", histogram_prob_kde)
            # print("histogram_mean", histogram_mean)
            # print("histogram_mean_data", histogram_mean_data_dependent)
            # for i in range(1, 5):
            #     print()
            #     print(f'Sample         : Moment {i} value is {n_weighted_moment(np.array(x), np.array([1 for _ in x]), i)}')
            #     print(f'Mean           : Moment {i} value is {n_weighted_moment(np.array(histogram_mean), np.array(histogram_prob), i)}')
            #     print(f'Filtered sample: Moment {i} value is {n_weighted_moment(np.array(histogram_mean_data_dependent), np.array(histogram_prob), i)}')
            #     print(f'Filtered (kde) : Moment {i} value is {n_weighted_moment(np.array(histogram_mean_data_dependent), np.array(histogram_prob_kde), i)}')
                
            #     print()
            # # print((np.array(histogram_prob) * np.array(histogram_mean)).sum())
            # # print(np.array(xx).mean())


            column_info[idx] = (bins, _binner_dict, data.columns[idx])


            # assertion
            _sum = 0
            for _attribute in _discretized_attributes_group[-1]:
                _sum += probs[_attribute]
            if(len(_discretized_attributes_group[-1]) >= 2):
                assert abs(_sum - 1 ) <= 0.1, "Error in calculating probabilities of discretized variables. Computed value is "  + str(abs(_sum - 1 ))  + " in group " + str(_discretized_attributes_group[-1])


    if(verbose):
        print("\nvariable to attribute map ->")    
        print(variable_attribute_map)
        print("\nattribute to variable map ->")
        print(attribute_variable_map)
        print("\n\n\n")
        
    
    return attributes, sensitive_attributes, probs, np.array(calculated_weights), variable_attribute_map, column_info

        


def get_sensitive_attibutes(known_sensitive_features, features):
        """ 
        Return sensitive attributes in appropriate format
        """

        # Extract new names of sensitive attributes
        _sensitive_attributes = {} # it is a map because each entry contains all one-hot encoded variables
        for _column in features:
            if("_" in _column and _column.split("_")[0] in known_sensitive_features):
                if(_column.split("_")[0] not in _sensitive_attributes):
                    _sensitive_attributes[_column.split("_")[0]] = [_column]
                else:
                    _sensitive_attributes[_column.split("_")[0]].append(_column)
            elif(_column in known_sensitive_features):
                if(_column not in _sensitive_attributes):
                    _sensitive_attributes[_column] = [_column]
                else:
                    _sensitive_attributes[_column].append(_column)


        # Finally make a 2d list
        sensitive_attributes = []
        for key in _sensitive_attributes:
            sensitive_attributes.append(_sensitive_attributes[key])


        return sensitive_attributes



from setuptools import setup

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
  name = 'justicia',
  packages = ['justicia'],
  version = 'v0.0.3',
  license='MIT',
  description = 'This library can be used formally verify machine learning models on multiple fairness definitions.',
  long_description=long_description,
  long_description_content_type='text/markdown',
  author = 'Bishwamittra Ghosh',
  author_email = 'bishwamittra.ghosh@gmail.com',
  url = 'https://github.com/meelgroup/justicia',
  download_url = 'https://github.com/meelgroup/justicia/archive/v0.0.3.tar.gz',
  keywords = ['ML fairness verification', 'Application of SSAT solvers'],   # Keywords that define your package best
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3.7',
  ],
)


from justicia import utils
from justicia.linear_classifier_wrap import linear_classifier_wrap_Wrapper
import itertools
from justicia.verifier import Fairness_verifier
from justicia.decision_tree_wrap import dtWrapper
import numpy as np
import time 
from itertools import chain

class Metric():

    def __init__(self, model, model_name, data, sensitive_attributes, neg_model=None, encoding="Enum", filename="sample.sdimacs", verbose=True, feature_threshold = 0, linear_continuous_data =  True, correlation = False, timeout=900):
        self.encoding = encoding
        self.model = model
        self.model_name = model_name
        self.data = data
        self.given_sensitive_attributes = sensitive_attributes
        self._filename = filename
        self._neg_model = neg_model
        self.verbose = verbose
        self.meta_sensitive_groups = None
        self._encode_correlation = correlation
        self._correlation_constraints = []
        self.timeout = timeout            

        
        # preprocessing
        start_time = time.time()
        self._get_required_params(feature_threshold, linear_continuous_data)
        self._compute()
        self.time_taken = time.time() - start_time
        
        
    
    def _get_required_params(self, feature_threshold, linear_continuous_data):
        
        
        if(self.model_name == "lr" or self.model_name == "svm-linear"):

            self._given_weights = self.model.coef_[0]
            
            # when data contains continuous features, call specilized statistics
            if(linear_continuous_data):
                self._attributes, self._sensitive_attributes, self._probs, self._given_weights, self._variable_attribute_map, self.column_info = utils.get_statistics_for_linear_classifier_wrap(self.data, self._given_weights, self.given_sensitive_attributes, verbose=self.verbose)
            else:
                self._attributes, self._sensitive_attributes, self._probs, self._variable_attribute_map, self.column_info = utils.get_statistics_from_df(self.data, self.given_sensitive_attributes)
            
            # consider important features only
            if(feature_threshold is not None):
                feature_importance = abs(self._given_weights)
                feature_importance = 100.0 * (feature_importance / feature_importance.max())
                threshold = np.percentile(np.array(feature_importance), float(feature_threshold))
                self._weights = []
                for i in range(len(self._given_weights)):
                    if(feature_importance[i] < threshold):
                        self._weights.append(0)
                    else:
                        self._weights.append(self._given_weights[i])
            else:
                self._weights =  self._given_weights  

            self._bias = self.model.intercept_[0]
            

            
                

            
            # get classifier
            lr = linear_classifier_wrap_Wrapper(weights=self._weights, attributes = self._attributes, sensitive_attributes = self._sensitive_attributes, bias=self._bias, verbose=self.verbose)
            self._num_attributes = lr.num_attributes
            self._attributes = lr.attributes
            self._classifier = lr.classifier
            self._sensitive_attributes = lr.sensitive_attributes
            self._auxiliary_variables = lr.auxiliary_variables

            
            if(self.verbose):
                print("Total number of variables in the formula:", self._num_attributes)
                print("Attribute variables:", self._attributes)
                print("Auxiliary variables:", self._auxiliary_variables)
                print("sensitive feature: ", self._sensitive_attributes)
                if(self.encoding != "Enum-correlation"):
                    print("\nprobabilities:", self._probs)
                
            if("efficient" in self.encoding):
                lr_neg = linear_classifier_wrap_Wrapper(weights=self._weights, attributes = self._attributes, sensitive_attributes = self._sensitive_attributes, bias=self._bias, negate=True, verbose=self.verbose)
                self._num_attributes_neg = lr_neg.num_attributes
                self._attributes_neg = lr_neg.attributes
                self._classifier_neg = lr_neg.classifier
                self._sensitive_attributes_neg = lr_neg.sensitive_attributes
                self._auxiliary_variables_neg = lr_neg.auxiliary_variables


        elif(self.model_name == "dt"):

            _sensitive_attributes = utils.get_sensitive_attibutes(self.given_sensitive_attributes, self.data.columns.to_list())
            dt_pos = dtWrapper(self.model,self.data.columns.tolist(), _sensitive_attributes, verbose=self.verbose)
            self._classifier = dt_pos.classifier
            self._num_attributes = dt_pos.num_attributes
            self._attributes = dt_pos.attributes
            self._sensitive_attributes = dt_pos.sensitive_attributes
            self._auxiliary_variables = dt_pos.auxiliary_variables
            self._variable_attribute_map = dt_pos.variable_attribute_map
            self._probs = dt_pos.compute_probability(self.data, verbose=self.verbose)

            if("correlation" in self.encoding):
                self._temp_model = dt_pos
                        
            
            if("efficient" in self.encoding):
                dt_neg = dtWrapper(self.model,self.data.columns.tolist(), _sensitive_attributes, negate=True, verbose=self.verbose)
                self._classifier_neg = dt_neg.classifier
                self._attributes_neg = dt_neg.attributes
                self._num_attributes_neg = dt_neg.num_attributes
                self._sensitive_attributes_neg = dt_neg.sensitive_attributes
                self._auxiliary_variables_neg = dt_neg.auxiliary_variables
            


        elif(self.model_name == "CNF"):
            self._attributes, self._sensitive_attributes, self._probs, self._variable_attribute_map, self.column_info = utils.get_statistics_from_df(self.data, self.given_sensitive_attributes, verbose=self.verbose)
            __classifier = self.model.get_selected_column_index()
            self._classifier = [[] for _ in __classifier]
            for idx in range(len(__classifier)):
                for (_var, _phase) in __classifier[idx]:
                    if(_phase >= 0):
                        self._classifier[idx].append(_var + 1)
                    else:
                        self._classifier[idx].append(-1 * (_var + 1))
            self._auxiliary_variables = []
            self._num_attributes = len(self._attributes + list(chain.from_iterable(self._sensitive_attributes)))

            assert self._num_attributes == len(self.data.columns)


            if(self.encoding == "Learn-efficient"):
                self.encoding = "Learn"

                
                
        else:
            print(self.model_name, "is not defined")
            raise ValueError
        
            

        



    

    

    def _compute(self):

        
        

        self.meta_sensitive_groups = []
        for _group in self._sensitive_attributes:
            if(len(_group)==1):
                _group = [_group[0], -1*_group[0]]
            self.meta_sensitive_groups.append(len(_group))

        min_value = None
        max_value = None
        self.execution_error = False
        self.most_favored_group = None
        self.least_favored_group = None
       
       
        
        
            
        if(self.encoding == "Enum-correlation"):
            
            max_value, min_value = self._run_Enum_correlation()

                 

        # Base encoding   
        elif(self.encoding == "Enum"):
            
            max_value, min_value = self._run_Enum()
        
            
        elif(self.encoding == "Learn-efficient"):
            
            max_value, min_value = self._run_Learn_efficient()
            

        elif(self.encoding == "Learn"):
            
            max_value, min_value = self._run_Learn()
        
        else:
            # print(self.encoding, "is not defined. Try RE-SSAT, ER-SSAT or ER-SSAT-efficient")
            # raise ValueError  
            pass

        if(self.execution_error == True):
            if(self.verbose):
                print("Execution error occured")
            self.statistical_parity_difference = None
            self.disparate_impact_ratio = None

            return

            

        self.statistical_parity_difference = max_value - min_value
        if(max_value ==  min_value):
            self.disparate_impact_ratio = 1
        elif(max_value == 0):
            self.disparate_impact_ratio = float("inf")
        else:
            self.disparate_impact_ratio = float(min_value/max_value)

    def _run_Enum(self, dependency_constraints=[]):
        # prepare for combination
        _sensitive_attributes = []
        for _group in self._sensitive_attributes:
            if(len(_group)==1):
                _group = [_group[0], -1*_group[0]]
            _sensitive_attributes.append(_group)
        
        # For each combination, update answer
        min_value = 1
        max_value = 0
        _combinations = list(itertools.product(*_sensitive_attributes))
        for configuration in _combinations:
            if(self.verbose):
                print("configuration: ", configuration, ":", self._get_group_from_configuration(configuration))
            fv = Fairness_verifier(timeout=float(self.timeout/len(_combinations)))
            fv.encoding_Enum_SSAT(self._classifier,self._attributes,self._sensitive_attributes,self._auxiliary_variables,self._probs, self._filename, dependency_constraints=dependency_constraints, sensitive_attributes_assignment=list(configuration), verbose=self.verbose)
            flag = fv.invoke_SSAT_solver(self._filename, verbose = self.verbose)
            
            
            if(flag == True):
                self.execution_error = True
                break

            if(min_value > fv.upper_bound):
                min_value = fv.upper_bound
                self.least_favored_group = self._get_group_from_configuration(configuration)
            if(max_value < fv.upper_bound):
                max_value = fv.upper_bound
                self.most_favored_group = self._get_group_from_configuration(configuration)

        return max_value, min_value

    def _get_group_from_configuration(self, configuration):
        result = []
        for var in configuration:
            if var > 0:
                if(self.model_name in ["lr", "CNF", "svm-linear"]):
                    result.append(self._variable_attribute_map[var])
                elif(self.model_name == "dt"):
                    (_feature, _threshold) = self._variable_attribute_map[var]
                    if(_threshold == 0.5):
                        result.append("not " + _feature)
                    else:
                        result.append(_feature + " <= " + str(_threshold))
                else:
                    raise ValueError
            else:
                if(self.model_name in ["lr", "CNF", "svm-linear"]):
                    result.append("not " + self._variable_attribute_map[-1 * var])
                elif(self.model_name == "dt"):
                    (_feature, _threshold) = self._variable_attribute_map[-1 * var]
                    if(_threshold == 0.5):
                        result.append(_feature)
                    else:
                        result.append(_feature + " > " + str(_threshold))
                else:
                    raise ValueError    

        return result

    def _run_Learn_efficient(self, dependency_constraints=[]):
        # maximum probability
        fv = Fairness_verifier(timeout=float(self.timeout/2))
        fv.encoding_Learn_SSAT(self._classifier,self._attributes,self._sensitive_attributes,self._auxiliary_variables,self._probs,self._filename, dependency_constraints=dependency_constraints, find_maximization = True, verbose = self.verbose)
        flag = fv.invoke_SSAT_solver(self._filename, find_maximization = True, verbose=self.verbose)
        max_value = fv.lower_bound
        self.most_favored_group = self._get_group(fv)

        if(flag == True):
            self.execution_error = True


        # minimum probability
        fv = Fairness_verifier(timeout=float(self.timeout/2))
        fv.encoding_Learn_SSAT(self._classifier_neg,self._attributes_neg,self._sensitive_attributes_neg,self._auxiliary_variables_neg,self._probs,self._filename, dependency_constraints=dependency_constraints, find_maximization = True, verbose = self.verbose)
        flag = fv.invoke_SSAT_solver(self._filename, find_maximization = False, verbose = self.verbose)
        if(flag == True):
            self.execution_error = True
                
        min_value = fv.upper_bound
        self.least_favored_group = self._get_group(fv)

        return max_value, min_value
    
    def _run_Learn(self, dependency_constraints=[]):
        # maximum probability
        fv = Fairness_verifier(timeout=float(self.timeout/2))
        fv.encoding_Learn_SSAT(self._classifier,self._attributes,self._sensitive_attributes,self._auxiliary_variables,self._probs,self._filename, dependency_constraints=dependency_constraints, find_maximization = True, verbose = self.verbose)
        flag = fv.invoke_SSAT_solver(self._filename, find_maximization = True, verbose=self.verbose)
        if(flag == True):
            self.execution_error = True

        max_value = fv.lower_bound
        self.most_favored_group = self._get_group(fv)


        # maximum probability by complementing the classifier
        fv = Fairness_verifier(timeout=float(self.timeout/2))
        fv.encoding_Learn_SSAT(self._classifier,self._attributes,self._sensitive_attributes,self._auxiliary_variables,self._probs,self._filename,  dependency_constraints=dependency_constraints, find_maximization = False, verbose = self.verbose)
        flag = fv.invoke_SSAT_solver(self._filename, find_maximization = False, verbose=self.verbose)
        if(flag == True):
            self.execution_error = True

        min_value = fv.upper_bound
        self.least_favored_group = self._get_group(fv)

        return max_value, min_value
        
    def _get_group(self, fairness_verifier):
        
        configuration = []
        for group in self._sensitive_attributes:
            if(len(group) == 1):
                if(group[0] in fairness_verifier.assignment_to_exists_variables):
                    configuration.append(group[0])
                else:
                    configuration.append(-1 * group[0])
            elif(len(group) > 1):
                reach = 0
                for var in group:
                    if(var in fairness_verifier.assignment_to_exists_variables):
                        configuration.append(var)
                        reach += 1
                assert reach == 1
            else:
                raise ValueError
        return self._get_group_from_configuration(configuration)
        

    def _get_df_mask(self, dominating_var, mask):
        # Each dominating var put constraints on the dataframe, which is captured by the mask.
        # When dominating var is negative, we simply calculate for positive var and finally complement it.
        
        negate_mask = False
        if(dominating_var < 0):
            # we need to negate the mask
            negate_mask = True
            dominating_var = -1 * dominating_var

        if(type(self._variable_attribute_map[dominating_var]) == tuple):
            if(self.model_name == "dt"):
                (_feature, _threshold) = self._variable_attribute_map[dominating_var]
                mask = mask & (self.data[_feature] <= _threshold)
            elif(self.model_name == "lr" or self.model_name == "svm-linear"):
                _feature, _, _threshold_1, _comparator_2, _threshold_2 = self._variable_attribute_map[dominating_var]
                if(_comparator_2 == "<"):
                    mask = mask & (self.data[_feature] >= _threshold_1) & (self.data[_feature] < _threshold_2)
                elif(_comparator_2 == "<="):
                    mask = mask & (self.data[_feature] >= _threshold_1) & (self.data[_feature] <= _threshold_2)
                else:
                    raise ValueError(_comparator_2)
            else:
                raise ValueError(self.model_name)
        else:
            mask = mask & (self.data[self._variable_attribute_map[dominating_var]] == 1)       

        if(negate_mask):
            mask = ~mask

        return mask   


    def _parameter_estimation_frequency(self, edges):
        # Dominating vars: list of nodes where a directed edge starts
        dominating_vars = list(set([a for (a,b) in edges]))            
        # dominating_vars should have all combinations
        edge_weights = {}
        for dominating_var in dominating_vars:

            mask = (True)

            if(isinstance(dominating_var, tuple)):
                # multuple parents
                for each_var in dominating_var:
                    mask = self._get_df_mask(each_var, mask)                        
                pass

            
            else:
                mask = self._get_df_mask(dominating_var, mask)
            
                    
            
            if(self.model_name == "dt"):
                marginal_probs = self._temp_model.compute_probability(self.data[mask], verbose=self.verbose)
            elif(self.model_name in ["lr", "CNF", "svm-linear"]):
                marginal_probs = utils.calculate_probs_linear_classifier_wrap(self.data[mask], self.column_info)
            else:
                raise ValueError
            
            for var in marginal_probs:
                edge_weights[(dominating_var,var)] = marginal_probs[var]
        
        return edge_weights
                

    def _run_Enum_correlation(self):
        _sensitive_attributes = []
        for _group in self._sensitive_attributes:
            if(len(_group)==1):
                _group = [_group[0], -1*_group[0]]
            _sensitive_attributes.append(_group)

        
        min_value = 1
        max_value = 0
        _combinations = list(itertools.product(*_sensitive_attributes))
        for configuration in _combinations:
            if(self.verbose):
                print("configuration: ", configuration, ":", self._get_group_from_configuration(configuration))
                # ":", [self._variable_attribute_map[var] if var > 0 else "not " + self._variable_attribute_map[-1 * var] for var in configuration]
            
            # recalculate probs according to conditional probabiities of sensitve attributes
            mask = (True)
            for _attribute in configuration:
                # if(_attribute in self._variable_attribute_map or -1 * _attribute in self._variable_attribute_map):

                #     if(self.model_name == "dt"):
                #         if(_attribute > 0):
                #             (_feature, _threshold) = self._variable_attribute_map[_attribute]
                #         else:
                #             (_feature, _threshold) = self._variable_attribute_map[-1 * _attribute]
                #         mask = mask & (self.data[_feature] <= _threshold)
                #     elif(self.model_name in ["lr", "svm-linear", "CNF"]):
                #         if(_attribute > 0):
                #             mask = mask & (self.data[self._variable_attribute_map[_attribute]] == 1)
                #         else:
                #             mask = mask & (self.data[self._variable_attribute_map[-1 * _attribute]] == 0)                                
                            
                #     else:
                #             raise ValueError()
                # else:
                #     raise ValueError(_attribute)

                mask = self._get_df_mask(_attribute, mask) 
            
                

            if(self.model_name == "dt"):
                self._probs = self._temp_model.compute_probability(self.data[mask], verbose=self.verbose)
            elif(self.model_name in ["lr", "CNF", "svm-linear"]):
                self._probs = utils.calculate_probs_linear_classifier_wrap(self.data[mask], self.column_info)
            else:
                raise ValueError
                
            if(self.verbose):
                print(self._probs)
            

            fv = Fairness_verifier(timeout=float(self.timeout/len(_combinations)))
            fv.encoding_Enum_SSAT(self._classifier,self._attributes,self._sensitive_attributes,self._auxiliary_variables,self._probs, self._filename, sensitive_attributes_assignment=list(configuration), verbose=self.verbose)
            flag = fv.invoke_SSAT_solver(self._filename, verbose = self.verbose)
            
            
            if(flag == True):
                self.execution_error = True
                max_value ==  min_value
                break

            if(min_value > fv.upper_bound):
                min_value = fv.upper_bound
                self.least_favored_group = self._get_group_from_configuration(configuration)
            if(max_value < fv.upper_bound):
                max_value = fv.upper_bound
                self.most_favored_group = self._get_group_from_configuration(configuration)

        return max_value, min_value






# wrapper for logistic regression
from pysat.pb import *
from pysat.solvers import Glucose3
from fractions import gcd
from functools import reduce
# from fairness_comparison.fairness.data.objects.ProcessedData import ProcessedData
# from fairness_comparison.fairness.data.objects.Adult import Adult
import pandas as pd
from sklearn.model_selection import train_test_split, KFold
import os
from sklearn.linear_model import LogisticRegression
from sklearn.svm import SVC
import sklearn.datasets
import numpy as np
from sklearn.datasets import load_iris
import justicia.utils as utils
from sklearn import metrics
from data.objects import titanic as titanic_
import numpy as np

class linear_classifier_wrap_Wrapper():

    # TODO
    # We are given a list of weight and a bias :  w = weights of features, b = bias
    # assumption: attributes are Boolean. 

    def _find_gcd(self, list):
        x = reduce(gcd, list)
        return x


    def __init__(self, weights, attributes, sensitive_attributes, bias, negate = False, verbose = True):
        self.weights = weights
        self.bias = bias
        self.classifier = None
        self.multiplier = 30
        self.num_attributes = len(weights)
        self.sensitive_attributes = sensitive_attributes
        self.attributes = attributes


        self._convert_weights_to_integers()
        self._get_cnf(negate)
        if(verbose):
            print("Expression: ", " ".join(map(str,self.weights)), "?=" , -1*self.bias)
            # print("clauses: ", self.classifier)
            
         
        

    def _convert_weights_to_integers(self):
        # convert w, bias to integer
        _max = abs(np.array(self.weights + [self.bias])).max()
        self.weights = [int(float(weight/_max)*self.multiplier) for weight in self.weights]
        self.bias = int(float(self.bias/_max)*self.multiplier)
        



    def _get_cnf(self, negate = False):
        # apply pseudo-Boolean encoding to get the CNF of the classifier
        lits = [i+1 for i in range(len(self.attributes) + len([_var for _group in self.sensitive_attributes for _var in _group]))]
        if(negate):
            cnf = PBEnc.atmost(lits = lits, weights = self.weights,  bound = -1 * self.bias - 1, top_id = self.num_attributes)
        else:
            cnf = PBEnc.atleast(lits = lits, weights = self.weights,  bound = -1 * self.bias, top_id = self.num_attributes)
        self.classifier = cnf.clauses
        self.auxiliary_variables = [i for i in range(self.num_attributes + 1, cnf.nv + 1)]
        self.num_attributes = cnf.nv



    def check_assignment(self):
        # check whether there is a satisfying assignment to the classifier. 
        g = Glucose3() 
        for clause in self.classifier:
            g.add_clause(clause)

        print("is SAT? ",g.solve())
        print("assignment: ",g.get_model())





# supporting functions for learning weights and bias


def init_iris():
    """ 
    Returns weights, bias, features (including sensitive features), and sensitive features
    """

    # loading dataset

    target = "target"
    dataset = load_iris()
    dataset[target] = np.where(dataset[target]==2, 0, dataset[target])

    # get df

    data_df = utils.sklearn_to_df(dataset)


    # discretize
    data =  utils.get_discretized_df(data_df, columns_to_discretize=data_df.columns.to_list())

    # get X,y
    X = data.drop(['target'], axis=1)
    y = data['target']

    
    # one-hot
    X = utils.get_one_hot_encoded_df(X,X.columns.to_list())

    # split into train_test
    X_train, X_test, y_train, y_test =  train_test_split(X,y, test_size=0.3, random_state=0)


    known_sensitive_attributes = [['sepal length (cm)_1']]
    attributes, sensitive_attributes, probs = utils.get_statistics_from_df(X_train, known_sensitive_attributes)



    #  For linear classifier, we use Logistic regression model of sklearn
    clf = LogisticRegression(random_state=0)
    clf = clf.fit(X_train, y_train)


    print("\nFeatures: ", X_train.columns.to_list())
    print("\nWeights: ", clf.coef_)
    print("\nBias:", clf.intercept_[0])
    assert len(clf.coef_[0]) == len(X_train.columns), "Error: wrong dimension of features and weights"


    print("Train Accuracy Score: ", clf.score(X_train, y_train), "positive ratio: ",y_train.mean())
    print("Test Accuracy Score: ", clf.score(X_test, y_test), "positive ratio: ",y_test.mean())


    return clf.coef_[0], clf.intercept_[0], attributes, sensitive_attributes, probs



# process adult dataset




def init_synthetic():
    filename = "data/sample.csv"

    if( os.path.isfile(filename)): 
        dataframe = pd.read_csv(filename)
    else:
        cols = 10
        rows = 200

        matrix = np.random.randint(2, size=(rows, cols))
        dataframe = pd.DataFrame.from_records(matrix)
        dataframe.columns = ['col-' + str(i) for i in range(cols-1)] + ['target']
        dataframe.to_csv(filename, index=False)


    # get X,y
    X = dataframe.drop(['target'], axis=1)
    y = dataframe['target']


    # one-hot
    X = utils.get_one_hot_encoded_df(X,X.columns.to_list())

    # split into train_test
    X_train, X_test, y_train, y_test =  train_test_split(X,y, test_size=0.3, random_state=0)


    known_sensitive_attributes = ['col-0']    
    attributes, sensitive_attributes, probs = utils.get_statistics_from_df(X_train, known_sensitive_attributes)



    #  For linear classifier, we use Logistic regression model of sklearn
    clf = LogisticRegression(random_state=0)
    clf = clf.fit(X_train, y_train)


    print("\nFeatures: ", X_train.columns.to_list())
    print("\nWeights: ", clf.coef_)
    print("\nBias:", clf.intercept_[0])
    assert len(clf.coef_[0]) == len(X_train.columns), "Error: wrong dimension of features and weights"
    

    # print("Train Accuracy Score: ", clf.score(X_train, y_train), "positive ratio: ",y_train.mean())
    # print("Test Accuracy Score: ", clf.score(X_test, y_test), "positive ratio: ",y_test.mean())
    predict_train = clf.predict(X_train)
    predict_test = clf.predict(X_test)

    
    print("Train accuracy:",metrics.accuracy_score(y_train, predict_train), "positive ratio: ",y_train.mean())
    print("Test accuracy:",metrics.accuracy_score(y_test, predict_test), "positive ratio: ",y_test.mean())
    print("Train set positive prediction",predict_train.mean())
    print("Test set positive prediction",predict_test.mean())
    print()
    

    return clf.coef_[0], clf.intercept_[0], attributes, sensitive_attributes, probs



def init(dataset, classifier = "lr", repaired=False, verbose = False, compute_equalized_odds = False):
    df = dataset.get_df(repaired=repaired)

    # discretize
    # df =  utils.get_discretized_df(df, columns_to_discretize=dataset.continuous_attributes)

    # get X,y
    X = df.drop(['target'], axis=1)
    y = df['target']

    # one-hot
    X = utils.get_one_hot_encoded_df(X,dataset.categorical_attributes)
    # X = utils.get_one_hot_encoded_df(X,X.columns.to_list())
    

    skf = KFold(n_splits=5, shuffle=True, random_state=10)
    skf.get_n_splits(X, y)

    X_trains = []
    y_trains = []
    X_tests = []
    y_tests = []
    clfs = []
    
    for train, test in skf.split(X, y):

        X_trains.append(X.iloc[train])
        y_trains.append(y.iloc[train])
        X_tests.append(X.iloc[test])
        y_tests.append(y.iloc[test])


        clf = None

        if(classifier == "lr"):
            #  For linear classifier, we use Logistic regression model of sklearn
            clf = LogisticRegression(class_weight='balanced', solver='liblinear', random_state=0)
        elif(classifier == "svm-linear"):
            clf = SVC(kernel="linear")
        else:
            raise ValueError(classifier)

        clf = clf.fit(X_trains[-1], y_trains[-1])
        clfs.append(clf)

        if(verbose):
            print("\nFeatures: ", X_trains[-1].columns.to_list())
            print("Number of features:", len(X_trains[-1].columns.to_list()))
            print("\nWeights: ", clf.coef_[0])
            print("\nBias:", clf.intercept_[0])
            assert len(clf.coef_[0]) == len(X_trains[-1].columns), "Error: wrong dimension of features and weights"


            print("Train Accuracy Score: ", clf.score(X_trains[-1], y_trains[-1]), "positive ratio: ",y_trains[-1].mean())
            print("Test Accuracy Score: ", clf.score(X_tests[-1], y_tests[-1]), "positive ratio: ",y_tests[-1].mean())

    
    if(compute_equalized_odds):
        return clfs, X_trains, X_tests, dataset.known_sensitive_attributes, y_trains, y_tests
    

    return clfs, X_trains, X_tests, dataset.known_sensitive_attributes











